// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// analyzeFuncs.C
// $Id: analyzeFuncs.cpp,v 1.34 2005/03/10 23:03:10 korpela Exp $
//

#define DO_SMOOTH

#include "config.h"

#include <stdio.h>
#include <iostream>
#include <stdlib.h>
#include <math.h>
#include <memory.h>

#include "sincos.h"
#include "util.h"
#include "s_util.h"
#ifdef BOINC_APP_GRAPHICS
#include "boinc_api.h"
#include "reduce.h"
#include "graphics_data.h"
#include "graphics_api.h"
#ifdef DYNAMIC_GRAPHICS
#include "graphics_lib.h"
#endif
#endif

#include "fft8g.h"
#include "seti.h"
#include "analyze.h"
#include "analyzeReport.h"
#include "gaussfit.h"
#include "spike.h"
#include "malloc_a.h"
#include "analyzeFuncs.h"
#include "analyzePoT.h"
#include "chirpfft.h"
#include "worker.h"
#include "filesys.h"
#ifdef BOINC_APP_GRAPHICS
#include "gdata.h"
#include "sah_gfx.h"
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define INVALID_CHIRP 2e+20

ChirpFftPair_t* ChirpFftPairs = NULL;

double ProgressUnitSize;
double progress;

// The main analysis function.  Args:
// state pointer to data, # of points, starting chirp/fftlen
//  Must be called with unchirped data;
//  this function modifies (chirps) the data in place
// swi  parsed WU header

int seti_analyze (ANALYSIS_STATE& state) {
  float* DataIn = state.savedWUData;
  int NumDataPoints = state.npoints;
  float* ChirpedData = NULL;
  float* WorkData = NULL;
  float* PowerSpectrum = NULL;
  int num_cfft                  = 0;
  int cfft_left;
  double MinChirpStep            = 0.0;
  float last_chirp = INVALID_CHIRP, chirprate;
  double progress_diff, progress_in_cfft;
  int retval=0;
  int icfft;
  int NumFfts, ifft, fftlen;
  int CurrentSub;
  int FftNum;
  unsigned long bitfield=swi.analysis_cfg.analysis_fft_lengths;
  unsigned long FftLen;

  // fields need by the ooura fft logic
  int * BitRevTab[MAX_NUM_FFTS];
  float * CoeffTab[MAX_NUM_FFTS];

  // Allocate data array and work area arrays.
  // See docs in fft8g.C for sizing guidelines for BitRevTab and CoeffTab.

  ChirpedData = state.data;
#if defined(_WIN32) && defined(_DEBUG)
  PowerSpectrum = (float*)_aligned_malloc((NumDataPoints * sizeof(float)), MEM_ALIGN);
  memset(PowerSpectrum, '\0', (NumDataPoints * sizeof(float)));
#else
  PowerSpectrum = (float*) calloc_a(NumDataPoints, sizeof(float), MEM_ALIGN);
#endif
  if(PowerSpectrum == NULL) SETIERROR(MALLOC_FAILED, "PowerSpectrum == NULL");

  FftNum=0;
  FftLen=1;
  while (bitfield != 0) {
    if (bitfield & 1) {
      swi.analysis_fft_lengths[FftNum]=FftLen;
#if defined(_WIN32) && defined(_DEBUG)
      BitRevTab[FftNum] = (int*)_aligned_malloc((3+(int)sqrt((float)swi.analysis_fft_lengths[FftNum]) * sizeof(int)), MEM_ALIGN);
      memset(BitRevTab[FftNum], '\0', (3+(int)sqrt((float)swi.analysis_fft_lengths[FftNum]) * sizeof(int)));
#else
      BitRevTab[FftNum] = (int*) calloc_a(3+(int)sqrt((float)swi.analysis_fft_lengths[FftNum]), sizeof(int), MEM_ALIGN);
#endif
      if(BitRevTab[FftNum] == NULL)  SETIERROR(MALLOC_FAILED, "BitRevTab[FftNum] == NULL");
      BitRevTab[FftNum][0] = 0;
      FftNum++;
    }
    FftLen*=2;
    bitfield>>=1;
  }
  swi.num_fft_lengths=FftNum;

  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
#if defined(_WIN32) && defined(_DEBUG)
      CoeffTab[FftNum] = (float*)_aligned_malloc(((swi.analysis_fft_lengths[FftNum]/2) * sizeof(float)), MEM_ALIGN);
      memset(CoeffTab[FftNum], '\0', ((swi.analysis_fft_lengths[FftNum]/2) * sizeof(float)));
#else
      CoeffTab[FftNum] = (float*) calloc_a(swi.analysis_fft_lengths[FftNum]/2, sizeof(float), MEM_ALIGN);
#endif
      if(CoeffTab[FftNum] == NULL) SETIERROR(MALLOC_FAILED, "CoeffTab[FftNum] == NULL");
  }

  // Smooth Baseline

#ifdef DO_SMOOTH
#ifdef BOINC_APP_GRAPHICS
  if (!nographics()) strcpy(sah_graphics->status, "Doing Baseline Smoothing");
#endif
  retval = v_BaseLineSmooth(
             DataIn, NumDataPoints, swi.analysis_cfg.bsmooth_boxcar_length, 
             swi.analysis_cfg.bsmooth_chunk_size
           );
  if (retval) SETIERROR(retval,"from v_BaseLineSmooth");
#endif

  // gernerate table of chirp/fft pairs (we may read table from file if testing)
  if(cfft_file != NULL)
    num_cfft = ReadCFftFile(&ChirpFftPairs, &MinChirpStep);
  else
    num_cfft = GenChirpFftPairs(&ChirpFftPairs, &MinChirpStep);
  if(num_cfft == MALLOC_FAILED) {
    SETIERROR(MALLOC_FAILED, "num_cfft == MALLOC_FAILED");
  }

  // Get together various values that we'll need to analyse power over time
  ComputePoTInfo(num_cfft, NumDataPoints);

  // used to calculate percent done
  //ProgressUnitSize = GetProgressUnitSize(NumDataPoints, num_cfft, swi);
  ProgressUnitSize = GetProgressUnitSize(NumDataPoints, num_cfft);

#ifdef DUMP_CHIRP
  fprintf(stderr, "size  = %d MinChirpStep = %f\n", num_cfft, MinChirpStep);
  for(icfft = 0; icfft < num_cfft; icfft++) {
    fprintf(stderr,"%f  %d\n", ChirpFftPairs[icfft].ChirpRate,
            ChirpFftPairs[icfft].FftLen);
  }
#endif

  // Loop through chirp/fft pairs - this is the top level analysis loop.
  for (icfft = state.icfft; icfft < num_cfft; icfft++) {
    fftlen    = ChirpFftPairs[icfft].FftLen;
    chirprate = ChirpFftPairs[icfft].ChirpRate;

    // avoid redundant chirping
    if(!VeryClose(chirprate, last_chirp, MinChirpStep/100.0)) {

#ifdef BOINC_APP_GRAPHICS
      if (!nographics()) strcpy(sah_graphics->status, "Chirping data");
#endif

      retval = v_ChirpData(
                 DataIn,
                 ChirpedData,
                 chirprate,
                 NumDataPoints,
                 swi.subband_sample_rate
               );

      if (retval) SETIERROR(retval, "from v_ChirpData()");

      progress += (float)(ProgressUnitSize * CHIRP_PROGRESS_FACTOR);
    }

    last_chirp = chirprate;

    // Process this FFT length.
    // For a given FFT length (at a given chirp), we construct
    // PowerSpectrum[] which is a "waterfall" array of power spectra,
    // each fftlen long on the frequency axis and sample time long
    // on the time axis.
    // As we go along, we check each spectra for spikes.

    state.icfft = icfft;     // update analysis state

    // Find index into FFT length table for the current
    // FFT length.  This will be the same index needed
    // for ooura's coeffecient and bit reverse tables.
    for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
      if(swi.analysis_fft_lengths[FftNum] == fftlen) {
        break;
      }
    }

#ifdef BOINC_APP_GRAPHICS
      if (!nographics()) {
        sah_graphics->fft_info.chirp_rate = chirprate;
        sah_graphics->fft_info.fft_len = fftlen;
        strcpy(sah_graphics->status, "Computing Fast Fourier Transform");
      }
#endif

    // If PoT freq bin is non-negative, we are into PoT analysis
    // for this cfft pair and should not re-output an "ogh" line.
    if(state.PoT_freq_bin == -1) {
      retval = result_group_start();
      if (retval) SETIERROR(retval,"from result_group_start");
    }

    // Number of FFTs for this length
    NumFfts   = NumDataPoints / fftlen;

    // Work array for this length
#if defined(_WIN32) && defined(_DEBUG)
    WorkData = (float *)_aligned_malloc((fftlen * 2 * sizeof(float)), MEM_ALIGN);
    memset(WorkData, '\0', (fftlen * 2 * sizeof(float)));
#else
    WorkData = (float *)calloc_a(fftlen * 2, sizeof(float), MEM_ALIGN);
#endif
    if (WorkData == NULL) {
      SETIERROR(MALLOC_FAILED, "WorkData == NULL");
    }

#ifdef BOINC_APP_GRAPHICS
    GRAPH_BUFFER* gbp;
    if (!nographics()) {
      gbp = sah_graphics->get_generate_buffer(true);
      if (gbp) {
        gbp->rarray.init_data(fftlen, NumFfts);
      }
    }
#endif
    for (ifft = 0; ifft < NumFfts; ifft++) {

      CurrentSub = fftlen * 2 * ifft;
      memcpy( WorkData,
              &ChirpedData[CurrentSub],
              (int)(fftlen * 2 * sizeof(float))
            );

      // replace time with freq - ooura FFT
      cdft(fftlen*2, 1, WorkData, BitRevTab[FftNum], CoeffTab[FftNum]);

      // replace freq with power
      v_GetPowerSpectrum( WorkData,
                          &PowerSpectrum[CurrentSub / 2],
                          fftlen
                        );

      // any ETIs ?!
      // If PoT freq bin is non-negative, we are into PoT analysis
      // for this cfft pair and need not redo spike finding.
      if(state.PoT_freq_bin == -1) {
        retval = FindSpikes(
                   &PowerSpectrum[CurrentSub / 2],
                   fftlen,
                   ifft,
                   swi
                 );
        if (retval) SETIERROR(retval,"from FindSpikes");
      }

      //progress = ((float)icfft)/num_cfft + ((float)ifft)/(NumFfts*num_cfft);
      progress += (float)((ProgressUnitSize * SPIKE_PROGRESS_FACTOR) / NumFfts);
      if (progress > 1.0f) progress = 1.0f; // prevent display of > 100%
#ifdef BOINC_APP_GRAPHICS
      if (!nographics()) { 
        if (gbp) {
          gbp->rarray.add_source_row(PowerSpectrum+fftlen*ifft);
        }
        sah_graphics->local_progress = (((float)ifft+1)/NumFfts);
      }
#endif
      boinc_fraction_done(progress);
// jeffc
//fprintf(stderr, "S fft len %d  progress = %12.10f\n", fftlen, progress);
    } // loop through chirped data array

#ifdef BOINC_APP_GRAPHICS
    if (!nographics()) {
      if (gbp) {
        sah_graphics->generate_done(gbp);
      }
    }
#endif
    boinc_fraction_done(progress);
// jeffc
//fprintf(stderr, "Sdone fft len %d  progress = %12.10f\n", fftlen, progress);


    // Analyze Power over Time.  May return quickly if this FFT
    // length and/or this WUs slew rate places the data block
    // outside PoT analysis limits.
    retval = analyze_pot(PowerSpectrum, NumDataPoints, fftlen);
    if (retval) SETIERROR(retval,"from analyze_pot");

#ifdef BOINC_APP_GRAPHICS
    // switch the display back to "best of" signals
    //
    if (!nographics()) {
      sah_graphics->gi.copy(best_gauss, true);
      sah_graphics->pi.copy(best_pulse, true);
      sah_graphics->ti.copy(best_triplet, true);
    }
#endif
    // Force progress to 100% before calling result_group_end() to store
    //  100% in state file so it will survive exit & relaunch
    if (icfft == (num_cfft-1)) {
      progress = 1;
      boinc_fraction_done(progress);
// jeffc
//fprintf(stderr, "DONE fft len %d  progress = %12.10f\n", fftlen, progress);

    }
    retval = checkpoint();
    if (retval) SETIERROR(retval,"from checkpoint() in seti_analyse()");

#if defined(_WIN32) && defined(_DEBUG)
    _aligned_free(WorkData);
#else
    free_a(WorkData);
#endif
    WorkData = NULL;

    // Progress adjustment - start checking late in the game (90%).
    // If progress_in_cfft is more than 1 cfft pair ahead of
    // progress, add the diff divided by the number of cfft's we have left.
    if((progress_in_cfft = (float)icfft/(float)num_cfft) > 0.9) {
      progress_diff = progress_in_cfft - progress;
      if(progress_diff > 1.0f/(float)num_cfft) {
        cfft_left = num_cfft-icfft > 0 ? num_cfft-icfft : 1;
        progress += progress_diff/cfft_left;
      }
    }

  } // loop over chirp/fftlen paris

  // Return the "best of" signals.  This may include duplicates of
  // already reported interesting signals.
  if(best_spike->score) {
    	retval = outfile.printf("%s", best_spike->s.print_xml(0,0,1,"best_spike").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best spike) in seti_analyze()");
	}

  }
  if(best_gauss->score) {
    	retval = outfile.printf("%s", best_gauss->g.print_xml(0,0,1,"best_gaussian").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best gaussian) in seti_analyze()");
	}
  }
  if(best_pulse->score) {
    	retval = outfile.printf("%s", best_pulse->p.print_xml(0,0,1,"best_pulse").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best pulse) in seti_analyze()");
	}
  }
  if(best_triplet->score) {
    	retval = outfile.printf("%s", best_triplet->t.print_xml(0,0,1,"best_triplet").c_str());
	if (retval < 0) {
  		SETIERROR(WRITE_FAILED,"from outfile.printf (best triplet) in seti_analyze()");
	}
  }

#ifdef BOINC_APP_GRAPHICS
  if (!nographics()) strcpy(sah_graphics->status, "Work unit done");
#endif

#if defined(_WIN32) && defined(_DEBUG)

  if (PowerSpectrum) _aligned_free(PowerSpectrum);
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (BitRevTab[FftNum]) _aligned_free(BitRevTab[FftNum]);
  }
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (CoeffTab[FftNum]) _aligned_free(CoeffTab[FftNum]);
  }
  if (WorkData) _aligned_free(WorkData);

#else

  if (PowerSpectrum) free_a(PowerSpectrum);
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (BitRevTab[FftNum]) free_a(BitRevTab[FftNum]);
  }
  for(FftNum = 0; FftNum < swi.num_fft_lengths; FftNum++) {
    if (CoeffTab[FftNum]) free_a(CoeffTab[FftNum]);
  }
  if (WorkData) free_a(WorkData);

#endif

  if (ChirpFftPairs) free(ChirpFftPairs);

  // jeffc
  //retval = outfile.flush();
  retval = checkpoint();		// force a final checkpoint
  xml_indent(-2);
  outfile.printf("</result>");
  outfile.close();
  //if (retval) SETIERROR(WRITE_FAILED,"from outfile.fflush in seti_analyze()");

  return retval;
}


int v_BaseLineSmooth(
  float* DataIn,
  int NumDataPoints,
  int BoxCarLength,
  int NumPointsInChunk
) {

  // We use a sliding boxcar method for baseline smoothing.  Input data
  // is the time domain.  It is transformed (using a separate array)
  // into the frequency domain.  After baseline smoothing is done
  // in the frequency domain, it is transformed back into the time domain.

  int h, i, j, k, i_minus_1;
  float* DataInChunk;
  float* PowerSpectrum;
  float Total, LocalMean, ScaleFactor;
  int NumTimeChunks, TimeChunk, Endpoint;

  int * BitRevTab = NULL;
  float * CoeffTab = NULL;

  NumTimeChunks = NumDataPoints / NumPointsInChunk;

#if defined(_WIN32) && defined(_DEBUG)
  PowerSpectrum = (float*) _aligned_malloc((NumPointsInChunk * sizeof(float)), MEM_ALIGN);
  memset(PowerSpectrum, '\0', (NumPointsInChunk * sizeof(float)));
#else
  PowerSpectrum = (float*) calloc_a(NumPointsInChunk, sizeof(float), MEM_ALIGN);
#endif
  if (PowerSpectrum == NULL) {
    printf("Could not allocate Power Spectrum array in v_BaseLineSmooth()\n");
    exit(1);
  }

#if defined(_WIN32) && defined(_DEBUG)
  BitRevTab = (int*) _aligned_malloc((3+(int)sqrt((float)NumPointsInChunk/2) * sizeof(int)), MEM_ALIGN);
  memset(BitRevTab, '\0', (3+(int)sqrt((float)NumPointsInChunk/2) * sizeof(int)));
#else
  BitRevTab = (int*) calloc_a(3+(int)sqrt((float)NumPointsInChunk/2), sizeof(int), MEM_ALIGN);
#endif
  if(BitRevTab == NULL) return MALLOC_FAILED;

#if defined(_WIN32) && defined(_DEBUG)
  CoeffTab = (float*) _aligned_malloc((NumPointsInChunk/2 * sizeof(float)), MEM_ALIGN);
  memset(CoeffTab, '\0', (NumPointsInChunk/2 * sizeof(float)));
#else
  CoeffTab = (float*) calloc_a(NumPointsInChunk/2, sizeof(float), MEM_ALIGN);
#endif
  if(CoeffTab == NULL) return MALLOC_FAILED;

  // flag to tell cdft() to init it's work areas
  BitRevTab[0] = 0;

  for (TimeChunk = 0; TimeChunk < NumTimeChunks; TimeChunk++) {
#ifdef BOINC_APP_GRAPHICS
    if (!nographics()) sah_graphics->local_progress = (((float)TimeChunk)/NumTimeChunks);
#endif

    DataInChunk = &DataIn[TimeChunk*NumPointsInChunk*2];

    // transform to freq
    cdft(NumPointsInChunk*2, 1, DataInChunk, BitRevTab, CoeffTab);

    v_GetPowerSpectrum(
      DataInChunk, PowerSpectrum, NumPointsInChunk
    );

    // Begin: normalize in freq. domain via sliding boxcar

    Endpoint = NumPointsInChunk / 2;

    i = Endpoint;         // start i at lowest negative freq;
    // this is low point in first boxcar
    j = i + BoxCarLength / 2;        // start j midpoint in first boxcar
    k = i + BoxCarLength;            // start k at end of first boxcar;
    // thus a boxcar is i----j----k
    Total = 0;
    for (h = i; h < k; h++) {            // Get mean for first boxcar
      Total += PowerSpectrum[h];
    }
    LocalMean = Total / BoxCarLength;
    ScaleFactor = (float)(sqrt(LocalMean / 2));

    // normalize 1st half of 1st boxcar
    for (h = i; h < j; h++) {
      DataInChunk[h*2]     /= ScaleFactor;
      DataInChunk[h*2+1] /= ScaleFactor;
    }

    for (; k != Endpoint; i++, j++, k++) {  // sliding boxcar
      if (k == NumPointsInChunk) {    // take care of wrapping
        k = 0;
      }
      if (j == NumPointsInChunk) {
        j = 0;
      }
      if (i == NumPointsInChunk) {
        i = 0;
        i_minus_1 = NumPointsInChunk -1;
      } else {
        i_minus_1 = i - 1;
      }

      LocalMean = LocalMean
                  - PowerSpectrum[i_minus_1] / BoxCarLength
                  + PowerSpectrum[k] / BoxCarLength;
      ScaleFactor = (float)(sqrt(LocalMean / 2));

      DataInChunk[j*2]   /= ScaleFactor;
      DataInChunk[j*2+1] /= ScaleFactor;

    } // End sliding boxcar

    // normalize final half of final boxcar
    for (h = j; h < k; h++) {
      DataInChunk[h*2]     /= ScaleFactor;
      DataInChunk[h*2+1] /= ScaleFactor;
    }

    // End: normalize in freq. domain via sliding boxcar

    // transform back to time
    cdft(NumPointsInChunk*2, -1, DataInChunk, BitRevTab, CoeffTab);

    // return powers to normal
    for (i = 0; i < NumPointsInChunk * 2; i++) {
      DataInChunk[i] /= NumPointsInChunk;
    }
  }

#if defined(_WIN32) && defined(_DEBUG)

  if (PowerSpectrum) _aligned_free(PowerSpectrum);
  if (BitRevTab) _aligned_free(BitRevTab);
  if (CoeffTab) _aligned_free(CoeffTab);

#else

  if (PowerSpectrum) free_a(PowerSpectrum);
  if (BitRevTab) free_a(BitRevTab);
  if (CoeffTab) free_a(CoeffTab);

#endif
  return 0;
}


void v_GetPowerSpectrum(
  float* FreqData,
  float* PowerSpectrum,
  int NumDataPoints
) {
  int i, j;

  for (i = 0, j = 0; i < NumDataPoints * 2; i += 2, j++) {
    PowerSpectrum[j] = FreqData[i] * FreqData[i]
                       + FreqData[i+1] * FreqData[i+1];
  }

}

// chirp_rate is in Hz per second

#ifndef USE_INTEL_OPT_CODE
int v_ChirpData(
  float* fp_DataArray,
  float* fp_ChirpDataArray,
  float chirp_rate,
  int  ul_NumDataPoints,
  double sample_rate
) {
  int i, j;
  float c, d, real, imag;
  double time, ang;

  if (chirp_rate == 0.0) {
    memcpy(fp_ChirpDataArray,
           fp_DataArray,
           (int)ul_NumDataPoints * 2 * sizeof(float)
          );    // NOTE INT CAST
  } else {
    for (i = 0, j = 0; i < ul_NumDataPoints * 2; i += 2, j++) {
      time = j/sample_rate;
      ang = M_PI*2*chirp_rate*time*time;
      sincosf(ang,&d,&c);

      // Sometimes chirping is done in place.
      // We don't want to overwrite data prematurely.
      real = fp_DataArray[i] * c - fp_DataArray[i+1] * d;
      imag = fp_DataArray[i] * d + fp_DataArray[i+1] * c;
      fp_ChirpDataArray[i] = real;
      fp_ChirpDataArray[i+1] = imag;
    }
  }
  return 0;
}
#endif // USE_INTEL_OPT_CODE
