// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// AlignedMalloc.c
// $Id: malloc_a.cpp,v 1.6 2004/07/02 21:21:08 korpela Exp $
#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "malloc_a.h"

// for processor portability (e.g. 32/64/etc, bit processors)
typedef size_t PointerArith;

// malloc_a and free_a are used to allocate one memory block

// Allocates memory on N-byte boundary
//
//
// Layout in memory of the pointers and buffer:
//
//  +-- malloc'd memory pointer
//  |
//  |          +-- save malloc'd memory pointer here for use with free_a()
//  V          V
//  +--------+---+------------------- ...
//  | unused |ptr|  aligned buffer
//  +--------+---+------------------- ...
//               ^
//               |
//               +-- aligned memory pointer returned from malloc_a()
//
//
//
void *malloc_a(size_t size, size_t alignment) {
  PointerArith byteAlignment;
  void *pmem;
  void *palignedMem;
  void **pp;

  // ensure byteAlignment is positive (if alignment is 0, make it 1)
  if (alignment < 1) {
    alignment = 1;
  }
  byteAlignment = alignment - 1;

  pmem = (void *) malloc(size + byteAlignment + sizeof(void *));    // allocate memory
  if (pmem == NULL) return NULL;

  // align memory on N byte boundary
  palignedMem = (void *) (((PointerArith) pmem + byteAlignment + sizeof(void *)) & ~(byteAlignment));

  pp = (void **) palignedMem;
  pp[-1] = pmem;    // store original address

  return(palignedMem);            // return aligned memory
}

//Frees memory that was allocated with malloc_a
void free_a(void *palignedMem) {
  void **pp;

  if (palignedMem == NULL) return;

  pp = (void **) palignedMem;

  //pp[-1] pts to original address of malloc'd memory
  free(pp[-1]);
}

void *calloc_a(size_t size, size_t nitems, size_t alignment) {
  void* p = malloc_a(size*nitems, alignment);
  if (p)
    memset(p, 0, size*nitems);
  return p;
}

